from bink.story import Story, story_from_file
import unittest


class StoryTestCase(unittest.TestCase):
    def test_oneline(self):
        story = story_from_file("inkfiles/oneline.ink.json")
        self.assertTrue(story.can_continue())
        self.assertEqual(story.cont(), "Line.\n")

    def test_load_save(self):
        """Test save_state and load_state functionality."""
        # Create a story and get initial text
        story = story_from_file("inkfiles/runtime/load-save.ink.json")

        # Continue to get all initial text
        lines = story.continue_maximally()

        # Check first line
        self.assertEqual(lines, "We arrived into London at 9.45pm exactly.\n")

        # Save the game state
        save_string = story.save_state()

        print(f"Save state: {save_string}")

        # Free the current story and create a new one
        del story

        story = story_from_file("inkfiles/runtime/load-save.ink.json")

        # Load the saved state
        story.load_state(save_string)

        # Choose first choice
        story.choose_choice_index(0)

        # Continue to get the next text
        lines = story.continue_maximally()

        # The text should contain both lines we expect
        self.assertIn("\"There is not a moment to lose!\" I declared.", lines)
        self.assertIn("We hurried home to Savile Row as fast as we could.", lines)

        # Check that we are at the end
        self.assertFalse(story.can_continue())

        # Check that there are no more choices
        choices = story.get_current_choices()
        self.assertEqual(len(choices), 0)

    def test_the_intercept(self):
        story = story_from_file("inkfiles/TheIntercept.ink.json")
        self.assertTrue(story.can_continue())

        end = False

        while not end:
            for line in story:
                print(line)  # Assuming 'line' is a byte-like object

            # Obtain and print choices
            choices = story.choices
            print(f"Num. choices: {len(choices)}\n")

            if choices:
                for i, text in enumerate(choices):
                    print(f"{i + 1}. {text}")

                # Always choose the first option
                story.choose_choice_index(0)
            else:
                end = True

        print("Story ended ok.")


if __name__ == '__main__':
    unittest.main()



# pylint: disable=E1101, C0116

"""Handle Ink Story."""
import ctypes
from bink.choices import Choices
from bink.tags import Tags
from bink import LIB, BINK_OK


class Story:
    """Story is the entry point of the Blade Ink lib."""
    def __init__(self, story_string: str):
        err_msg = ctypes.c_char_p()
        story = ctypes.c_void_p()
        ret = LIB.bink_story_new(
            ctypes.byref(story),
            story_string.encode('utf-8'),
            ctypes.byref(err_msg))
        if ret != BINK_OK:
            err = err_msg.value.decode('utf-8')
            LIB.bink_cstring_free(err_msg)
            raise RuntimeError(err)

        self._story = story

    def __next__(self):
        if not self.can_continue():
            raise StopIteration

        return self.cont()

    def __iter__(self):
        return self

    @property
    def choices(self):
        return self.get_current_choices()

    @property
    def tags(self):
        return self.get_current_tags()

    def can_continue(self):
        can_continue = ctypes.c_bool()
        ret = LIB.bink_story_can_continue(
            self._story, ctypes.byref(can_continue))

        if ret != BINK_OK:
            raise RuntimeError("Error in can_continue")

        return can_continue.value

    def cont(self) -> str:
        err_msg = ctypes.c_char_p()
        line = ctypes.c_char_p()
        ret = LIB.bink_story_cont(
            self._story,
            ctypes.byref(line),
            ctypes.byref(err_msg))

        if ret != BINK_OK:
            err = err_msg.value.decode('utf-8')
            LIB.bink_cstring_free(err_msg)
            raise RuntimeError(err)

        result = line.value.decode('utf-8')
        LIB.bink_cstring_free(line)

        return result

    def continue_maximally(self) -> str:
        err_msg = ctypes.c_char_p()
        line = ctypes.c_char_p()
        ret = LIB.bink_story_continue_maximally(
            self._story, ctypes.byref(line), ctypes.byref(err_msg))

        if ret != BINK_OK:
            err = err_msg.value.decode('utf-8')
            LIB.bink_cstring_free(err_msg)
            raise RuntimeError(err)

        result = line.value.decode('utf-8')
        LIB.bink_cstring_free(line)

        return result

    def get_current_choices(self) -> Choices:
        choices = ctypes.c_void_p()
        choice_count = ctypes.c_int()
        ret = LIB.bink_story_get_current_choices(
            self._story, ctypes.byref(choices), ctypes.byref(choice_count))

        if ret != BINK_OK:
            raise RuntimeError("Error getting current choices")

        choices = Choices(choices, choice_count.value)

        return choices

    def choose_choice_index(self, choice_index: int):
        """Chooses the `Choice` from the
        `currentChoices` list with the given index. Internally, this
        sets the current content path to what the
        `Choice` points to, ready to continue story evaluation."""
        err_msg = ctypes.c_char_p()
        cidx = ctypes.c_int(choice_index)
        ret = LIB.bink_story_choose_choice_index(
            self._story, cidx, ctypes.byref(err_msg))

        if ret != BINK_OK:
            err = err_msg.value.decode('utf-8')
            LIB.bink_cstring_free(err_msg)
            raise RuntimeError(err)

    def get_current_tags(self) -> Tags:
        tags = ctypes.c_void_p()
        tag_count = ctypes.c_int()
        ret = LIB.bink_story_get_current_tags(
            self._story, ctypes.byref(tags), ctypes.byref(tag_count))

        if ret != BINK_OK:
            raise RuntimeError("Error getting current tags")

        tags = Tags(tags, tag_count.value)

        return tags

    def choose_path_string(self, path: str):
        err_msg = ctypes.c_char_p()
        ret = LIB.bink_story_choose_path_string(
            self._story,
            path.encode('utf-8'),
            ctypes.byref(err_msg))
        if ret != BINK_OK:
            err = err_msg.value.decode('utf-8')
            LIB.bink_cstring_free(err_msg)
            raise RuntimeError(err)

    def save_state(self) -> str:
        """Saves the current state of the story and returns it as a string.
        The returned state can be loaded later using load_state()."""
        err_msg = ctypes.c_char_p()
        save_string = ctypes.c_char_p()
        ret = LIB.bink_story_save_state(
            self._story,
            ctypes.byref(save_string),
            ctypes.byref(err_msg))

        if ret != BINK_OK:
            err = err_msg.value.decode('utf-8')
            LIB.bink_cstring_free(err_msg)
            raise RuntimeError(err)

        result = save_string.value.decode('utf-8')
        LIB.bink_cstring_free(save_string)

        return result

    def load_state(self, save_state: str):
        """Loads a previously saved state into the story.
        This allows resuming the story from a saved point."""
        err_msg = ctypes.c_char_p()
        ret = LIB.bink_story_load_state(
            self._story,
            save_state.encode('utf-8'),
            ctypes.byref(err_msg))

        if ret != BINK_OK:
            err = err_msg.value.decode('utf-8')
            LIB.bink_cstring_free(err_msg)
            raise RuntimeError(err)

    def __del__(self):
        LIB.bink_story_free(self._story)


def story_from_file(story_file: str):
    with open(story_file, 'r', encoding='utf-8') as file:
        content = file.read()
        return Story(content)



"""Load the Bink C library."""
import os
import platform
import ctypes.util

def _load_library():
    # First, look for a bundled bink shared object on the lib folder.
    system = platform.system()
    library_name = None
    if system == 'Windows':
        library_name = 'bink.dll'
    elif system == 'Darwin':
        library_name = 'libbink.dylib'
    else:
        library_name = 'libbink.so'

    arch = "x86_64/"

    if platform.machine() == "arm64":
        arch = "arm64/"

    _filename = os.path.join(os.path.dirname(__file__), 'native/' + arch + library_name)

    # If no bundled shared object is found, look for a system-wide installed one.
    if not os.path.exists(_filename):
        # on windows all ctypes does when checking for the library
        # is to append .dll to the end and look for an exact match
        # within any entry in PATH.
        _filename = ctypes.util.find_library('bink')

        if _filename is None:
            if platform.system() == 'Windows':
                # Check current working directory for dll as ctypes fails to do so
                _filename = os.path.join(os.path.realpath('.'), "bink.dll")
            else:
                _filename = library_name

    try:
        #print("lib filename: ", _filename)
        lib = ctypes.CDLL(_filename)
    except (OSError, TypeError) as exc:
        lib = None
        raise RuntimeError('bink library not found') from exc
    return lib

LIB = _load_library()

LIB.bink_story_new.argtypes = [
    ctypes.POINTER(
        ctypes.c_void_p), ctypes.c_char_p, ctypes.POINTER(
            ctypes.c_char_p)]
LIB.bink_story_new.restype = ctypes.c_int

LIB.bink_story_can_continue.argtypes = [
    ctypes.c_void_p, ctypes.POINTER(ctypes.c_bool)]
LIB.bink_story_can_continue.restype = ctypes.c_int

BINK_OK = 0
BINK_FAIL = 1
BINK_FAIL_NULL_POINTER = 2



from bink.story import story_from_file
import unittest

class TagsInkTestCase(unittest.TestCase):
    def test_tags(self):
        story = story_from_file("inkfiles/tags.ink.json")
        self.assertTrue(story.can_continue())
        
        self.assertEqual("This is the content\n", story.cont())

        current_tags = story.get_current_tags()
        self.assertEqual(2, len(current_tags))
        self.assertEqual("author: Joe", current_tags[0])
        self.assertEqual("title: My Great Story", current_tags[1])

        story.choose_path_string("knot")
        self.assertEqual("Knot content\n", story.cont())
        current_tags = story.get_current_tags()
        self.assertEqual(1, len(current_tags))
        self.assertEqual("knot tag", current_tags[0])

        self.assertEqual("", story.cont())
        current_tags = story.get_current_tags()
        self.assertEqual("end of knot tag", current_tags[0])

if __name__ == '__main__':
    unittest.main()



# pylint: disable=E1101, R0903

"""Handle Ink tags."""
import ctypes
from bink import LIB, BINK_OK


class TagsIterator:
    """Iterator for tags."""
    def __init__(self, tags):
        self._tags = tags
        self._index = 0

    def __next__(self):
        if self._index >= len(self._tags):
            raise StopIteration

        self._index += 1
        return self._tags[self._index - 1]


class Tags:
    """Contains a list of tags."""
    def __init__(self, tags, c_len: int):
        self._tags = tags
        self._len = c_len

    def __len__(self) -> int:
        """Returns the number of tags."""
        return self._len

    def __bool__(self) -> bool:
        return self._len != 0

    def __iter__(self):
        return TagsIterator(self)

    def __getitem__(self, idx: int) -> str:
        """Returns the tag text"""

        if not isinstance(idx, int):
            raise TypeError

        if idx < 0 or idx >= self._len:
            raise IndexError

        return self.get(idx)

    def get(self, idx) -> str:
        """Returns the tag text."""
        tag = ctypes.c_char_p()
        ret = LIB.bink_tags_get(self._tags, idx, ctypes.byref(tag))

        if ret != BINK_OK:
            raise RuntimeError("Error getting tag, index out of bounds?")

        result = tag.value.decode('utf-8')
        LIB.bink_cstring_free(tag)

        return result

    def __del__(self):
        LIB.bink_tags_free(self._tags)



# pylint: disable=E1101, R0903
"""Handle Ink Choices."""
import ctypes
from bink import LIB, BINK_OK


class ChoicesIterator:
    """Iterator for choices."""
    def __init__(self, choices):
        self._choices = choices
        self._index = 0

    def __next__(self):
        if self._index >= len(self._choices):
            raise StopIteration

        self._index += 1
        return self._choices[self._index - 1]


class Choices:
    """List of story choices."""
    def __init__(self, choices, c_len: int):
        self._choices = choices
        self._len = c_len

    def __len__(self) -> int:
        """Returns the number of choices."""
        return self._len

    def __bool__(self) -> bool:
        return self._len != 0

    def __iter__(self):
        return ChoicesIterator(self)

    def __getitem__(self, idx: int) -> str:
        """Returns the choice text"""

        if not isinstance(idx, int):
            raise TypeError

        if idx < 0 or idx >= self._len:
            raise IndexError

        return self.get_text(idx)

    def get_text(self, idx) -> str:
        """Returns the choice text."""
        text = ctypes.c_char_p()
        ret = LIB.bink_choices_get_text(self._choices, idx, ctypes.byref(text))

        if ret != BINK_OK:
            raise RuntimeError(
                "Error getting choice text, index out of bounds?")

        result = text.value.decode('utf-8')
        LIB.bink_cstring_free(text)

        return result

    def __del__(self):
        LIB.bink_choices_free(self._choices)



"""Package definition."""


import sys
from setuptools import setup, Distribution
from wheel.bdist_wheel import bdist_wheel


class BinaryDistribution (Distribution):
    def has_ext_modules(self):
        return True


class BdistWheel(bdist_wheel):
    def get_tag(self):
        return ('py3', 'none') + super().get_tag()[2:]


def get_package_data():
    plat_name_idx = None
    # find --plat-name argument idx and get the next argument
    for i, arg in enumerate(sys.argv):
        if arg == '--plat-name':
            plat_name_idx = i + 1
            break

    # if --plat-name argument is present, return ['native/lib_name']
    if plat_name_idx:
        plat_name = sys.argv[plat_name_idx]
        if plat_name.startswith('macosx'):
            lib = 'libbink.dylib'
        elif plat_name.startswith('linux') or plat_name.startswith('manylinux'):
            lib = 'libbink.so'
        elif plat_name.startswith('win'):
            lib = 'bink.dll'
        else:
            raise RuntimeError('Unsupported platform: ' + plat_name)

        arch = "x86_64/"
        if "arm64" in plat_name or "aarch64" in plat_name:
            arch = "arm64/"

        lib = arch + lib

        return ['native/' + lib]

    # if it is not present, return ['native/*']
    return ['native/*']


setup(
    package_data={'bink': get_package_data()},
    distclass = BinaryDistribution,
    cmdclass = {
        'bdist_wheel': BdistWheel,
    },
)
